from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional

from melid.tailwind.color import generate_color_utilities
from melid.tailwind.configs.color import COLORS
from melid.tailwind.configs.size import SIZES
from melid.tailwind.configs.variants import VARIANTS
from melid.tailwind.css_ast import (
    Ast,
    AstNode,
    AtRoot,
    Context,
    Declaration,
    Rule,
    WalkAction,
)
from melid.tailwind.size import generate_size_utilities
from melid.tailwind.utils import generate_utilities_with_variants


@dataclass
class TailwindConfig:
    colors: Dict[str, str] = field(default_factory=lambda: COLORS)
    sizes: Dict[str, str] = field(default_factory=lambda: SIZES)
    variants: Dict[str, Dict[str, str]] = field(default_factory=lambda: VARIANTS)

    def __build_utilities_ast(self) -> List[AstNode]:
        """Generate AST nodes for the additional utilities."""
        utilities = {}

        utility_ast = []
        utility_ast.extend(
            generate_utilities_with_variants(
                {
                    **generate_color_utilities(self.colors),
                    **generate_size_utilities(self.sizes),
                },
                self.variants,
            )
        )

        for class_name, utility in utilities.items():
            utility_ast.append(Ast.rule(class_name, utility["declarations"]))

        return utility_ast

    @property
    def utilities(self):
        return self.__build_utilities_ast()

    def extend(
        self,
        colors: Dict[str, str] = None,
        sizes: Dict[str, str] = None,
        variants: Dict[str, Dict[str, str]] = None,
    ):
        if colors:
            self.colors.update(colors)

        if sizes:
            self.sizes.update(sizes)

        if variants:
            self.variants.update(variants)

        return self


class Tailwind:
    _INSTANCE = None

    # Build the main AST
    __initial_ast: List[AstNode] = []

    _initialized = False  # Flag to track if __init__ has already been executed

    def __new__(cls, *args, **kwargs):
        if cls._INSTANCE is None:
            cls._INSTANCE = super().__new__(cls)
        return cls._INSTANCE

    def __init__(self, config: Optional[TailwindConfig] = None):
        if not self._initialized:
            self._initialized = True  # Ensure this is only done once

            # Handle the config to avoid mutable default argument issues
            if config is None:
                config = TailwindConfig()

            self.__initial_ast.extend(config.utilities)

            # Walk the AST and modify it
            Ast.walk(self.__initial_ast, self.visit_function)

            self.__utility_map = self.build_utility_map(self.__initial_ast)

    def visit_function(
        self, node: AstNode, utils: Dict[str, Any]
    ) -> Optional[WalkAction]:
        """
        Traverse AST nodes and update utilities, variants, and colors.
        """
        if isinstance(node, Declaration):
            # Update color values or any other transformation here
            pass  # Customize as needed

        # Process variants (like hover, focus, etc.)
        elif isinstance(node, Context):
            for child in node.nodes:
                self.visit_function(child, utils)

        return WalkAction.CONTINUE

    def build_utility_map(self, ast: List[AstNode]) -> Dict[str, List[Declaration]]:
        """
        Build a utility map from the AST, mapping class names to their declarations.
        """
        utility_map: Dict[str, List[Declaration]] = {}

        def traverse(node: AstNode):
            if isinstance(node, Rule):
                if node.selector.startswith("."):
                    class_selector = node.selector[1:]
                    class_name = class_selector.replace("\\:", ":")
                    declarations = [
                        child for child in node.nodes if isinstance(child, Declaration)
                    ]
                    if declarations:
                        utility_map[class_name] = declarations
            elif isinstance(node, Context) or isinstance(node, AtRoot):
                for child in node.nodes:
                    traverse(child)

        for node in ast:
            traverse(node)

        return utility_map

    def process_class_names(
        self, class_names: str, utility_map: Dict[str, List[Declaration]]
    ) -> Dict[str, List[Declaration]]:
        """
        Process a string of Tailwind-like class names and return a dictionary of
        base and variant CSS declarations.
        """
        base_declarations: List[Declaration] = []
        variant_declarations: Dict[str, List[Declaration]] = {}

        classes = class_names.split()

        for cls in classes:
            if cls in utility_map:
                # Check if the class contains a variant (like hover:)
                if ":" in cls:
                    variant, base_class = cls.split(":")
                    if base_class in utility_map:
                        if variant not in variant_declarations:
                            variant_declarations[variant] = []
                        variant_declarations[variant].extend(utility_map[base_class])
                else:
                    base_declarations.extend(utility_map[cls])
            else:
                print(f"Warning: Class '{cls}' is not defined in the utility map.")

        return {
            "base": base_declarations,
            "variants": variant_declarations,
        }

    @classmethod
    def tw(cls, element: str, class_names: str) -> str:
        """
        Process class names and return corresponding CSS rules for an element,
        including variants like `hover`.
        """

        if not cls._INSTANCE:
            raise ValueError("Router has not been initialized")

        declarations = cls._INSTANCE.process_class_names(
            class_names, cls._INSTANCE.__utility_map
        )

        # Build the base CSS rule
        css_output = f"{element} {{\n"
        for decl in declarations["base"]:
            css_output += f"  {decl.property}: {decl.value};\n"
        css_output += "}\n"

        # Build the CSS rules for variants (e.g., hover, focus, etc.)
        for variant, decls in declarations["variants"].items():
            css_output += f"{element}:{variant} {{\n"
            for decl in decls:
                css_output += f"  {decl.property}: {decl.value};\n"
            css_output += "}\n"

        return css_output



import enum
import typing

from PyQt5 import QtCore, QtWidgets

from melid.utils import Size
from melid.widgets import Widget


class Direction(enum.Enum):
    COLUMN = QtWidgets.QBoxLayout.TopToBottom
    COLUMN_REVERSE = QtWidgets.QBoxLayout.BottomToTop
    ROW = QtWidgets.QBoxLayout.LeftToRight
    ROW_REVERSE = QtWidgets.QBoxLayout.RightToLeft


class Box(Widget):
    def __init__(
        self,
        child: typing.Optional[
            typing.Union[QtWidgets.QWidget, typing.List[QtWidgets.QWidget]]
        ] = None,
        direction: typing.Optional[Direction] = Direction.ROW,
        gap: typing.Optional[int] = 0,
        alignment: typing.Optional[
            typing.Union[QtCore.Qt.Alignment, QtCore.Qt.AlignmentFlag]
        ] = QtCore.Qt.AlignmentFlag.AlignTop
        | QtCore.Qt.AlignmentFlag.AlignLeft,
        size: typing.Optional[Size] = None,
        style: typing.Optional[str] = "",
    ):
        super().__init__()

        self.setGap(gap).setDirection(direction.value).setStyleSheet(style)

        if size:
            size.apply(self)

        if child:
            if isinstance(child, list) and len(child) > 0:
                self.addWidgets(child, alignment=alignment)

            if isinstance(child, QtWidgets.QWidget):
                self.addWidget(child, alignment=alignment)


class Scroll(Widget):
    def __init__(
        self,
        child: typing.Optional[
            typing.Union[QtWidgets.QWidget, typing.List[QtWidgets.QWidget]]
        ] = None,
        *args,
        **kwargs,
    ):
        super().__init__()

        self.__box = Box(child, *args, **kwargs)

        self.__scroll_area = QtWidgets.QScrollArea()
        self.__scroll_area.setStyleSheet(
            "QScrollArea {background-color: transparent; border: none;}"
        )
        self.__scroll_area.setContentsMargins(0, 0, 0, 0)
        self.__scroll_area.setWidget(self.__box)
        self.__scroll_area.setWidgetResizable(True)

        super().addWidget(self.__scroll_area)

    def onWindowResized(self, size: QtCore.QSize):
        print("w: %d, h: %d" % (size.width(), size.height()))

        # Size(
        #     width=size.width(),
        #     height=size.height(),
        # ).apply(self).apply(
        #     self.__box
        # ).apply(self.__scroll_area)

        # self.update()

    def addWidget(self, *args, **kwargs):
        return self.__box.addWidget(*args, **kwargs)



from typing import Any, Dict


def generate_color_utilities(colors: Dict[str, str]) -> Dict[str, Any]:
    utilities = {}

    for color_name, color_value in colors.items():
        # Background Color Utilities
        utilities[f"bg-{color_name}"] = {
            "declarations": [{"property": "background-color", "value": color_value}]
        }

        # Text Color Utilities
        utilities[f"text-{color_name}"] = {
            "declarations": [{"property": "color", "value": color_value}]
        }

        # Border Color Utilities
        utilities[f"border-{color_name}"] = {
            "declarations": [{"property": "border-color", "value": color_value}]
        }

    return utilities



from typing import Any, Dict, List

from melid.tailwind.css_ast import Ast, AstNode


def generate_ast(utilities: Dict[str, Any]) -> Dict[str, AstNode]:
    utility_ast = {}

    for class_name, utility in utilities.items():
        declarations = list(
            map(
                lambda declaration: Ast.decl(
                    declaration["property"], declaration["value"]
                ),
                utility["declarations"],
            )
        )
        rule_node = Ast.rule(f".{class_name}", declarations)
        utility_ast[class_name] = rule_node

    return utility_ast


def generate_utilities_with_variants(
    utilities: Dict[str, str], variants: Dict[str, Dict[str, str]]
) -> List[AstNode]:

    utility_ast = generate_ast(utilities)
    extended_ast: List[AstNode] = list(utility_ast.values())

    for variant_name, variant_detail in variants.items():
        variant_type = variant_detail["type"]
        if variant_type == "pseudo":
            for class_name, rule_node in utility_ast.items():
                # Escape colon in class names
                variant_class = f".{variant_name}\\:{class_name}"

                # Clone declarations
                declarations = [
                    Ast.decl(d.property, d.value, d.important) for d in rule_node.nodes
                ]
                variant_rule_node = Ast.rule(variant_class, declarations)
                extended_ast.append(variant_rule_node)

    return extended_ast



import enum
import types
import typing
from dataclasses import dataclass

from PyQt5 import QtCore, QtGui, QtWidgets

from melid.store import StatefulWidget
from melid.types import AppEvents


class Widget(QtWidgets.QGroupBox, StatefulWidget):

    STYLESHEET_PATH = ""
    STYLESHEET_TYPE = "CSS"

    def __init__(
        self,
        qt_window_resized: typing.Optional[QtCore.pyqtBoundSignal] = None,
    ):
        super().__init__()

        # if qt_window_resized:
        #     qt_window_resized.connect(self.onWindowResized)

        self.setObjectName(self.__class__.__name__)
        self.setContentsMargins(0, 0, 0, 0)

        # Apply stylesheets if a valid stylesheet path is provided
        if self.STYLESHEET_PATH:
            with open(self.STYLESHEET_PATH, "r") as stylesheet:
                self.setStyleSheet(
                    "QGroupBox {background-color: transparent; border: none; %s}"
                    % stylesheet.read()
                )

        # Initialize a QBoxLayout (Top to Bottom) for the widget
        self.__layout = QtWidgets.QBoxLayout(QtWidgets.QBoxLayout.TopToBottom, self)
        self.__layout.setContentsMargins(0, 0, 0, 0)

    def setGap(self, gap: int):
        self.__layout.setSpacing(gap)

        return self

    def setDirection(self, direction: QtWidgets.QBoxLayout.Direction):
        """
        Set the layout direction (e.g., TopToBottom, LeftToRight).
        """
        self.__layout.setDirection(direction)

        return self

    def addWidget(self, widget: QtWidgets.QWidget, *args, **kwargs):
        """
        Add a single widget to the layout.
        """

        kwargs.setdefault("stretch", widget.width())

        self.__layout.addWidget(widget, *args, **kwargs)

        return self

    def addWidgets(self, widgets: typing.List[QtWidgets.QWidget], *args, **kwargs):
        """
        Add multiple widgets progressively using QTimer to avoid freezing the UI.
        """

        def _addNextWidget():
            if len(widgets) > 0:
                # Add the next widget
                self.addWidget(widgets.pop(0), *args, **kwargs)

                QtWidgets.QApplication.processEvents()
            else:
                # Stop the timer once all widgets are added
                timer.stop()

        # Use a QTimer to add widgets progressively
        timer = QtCore.QTimer(self)
        timer.timeout.connect(_addNextWidget)
        timer.start()

        return self

    @QtCore.pyqtSlot(AppEvents.WINDOW_RESIZED.value)
    def onWindowResized(self, size: QtCore.QSize):
        pass

    def updateStyleSheet(self, style: str):
        self.setStyleSheet(self.styleSheet() + "\n%s" % style)


class Label(QtWidgets.QLabel, Widget):
    def __init__(self, text: str, style: str = "", *args, **kwargs):
        super(Label, self).__init__(*args, **kwargs)

        self.setObjectName(self.__class__.__name__)

        self.setText(text)
        self.setStyleSheet(style)


class Button(QtWidgets.QPushButton, Widget):
    def __init__(
        self,
        child: typing.Optional[typing.Union[str, QtWidgets.QWidget, QtGui.QIcon]] = "",
        on_click: typing.Optional[types.FunctionType] = None,
        style: typing.Optional[str] = "",
        *args,
        **kwargs,
    ):
        super(Button, self).__init__(*args, **kwargs)

        self.setObjectName(self.__class__.__name__)

        if child:
            if isinstance(child, str):
                self.setText(child)
            if isinstance(child, QtGui.QIcon):
                self.setIcon(child)

        if on_click:
            self.clicked.connect(lambda: on_click())

        self.setStyleSheet(style)
        self.setCursor(QtGui.QCursor(QtCore.Qt.PointingHandCursor))


class InputType(enum.Enum):
    TEXT = "TEXT"
    PASSWORD = "PASSWORD"
    EMAIL = "EMAIL"


class Input(QtWidgets.QLineEdit, Widget):
    def __init__(
        self,
        input_type: typing.Optional[InputType] = InputType.TEXT,
        placeholder: typing.Optional[str] = None,
        on_change: typing.Optional[typing.Callable] = None,
        style: typing.Optional[str] = "",
    ):
        super().__init__()

        if input_type == InputType.PASSWORD:
            self.setEchoMode(QtWidgets.QLineEdit.Password)

        if placeholder:
            self.setPlaceholderText(placeholder)

        if on_change:
            self.textChanged.connect(on_change)

        self.setObjectName(self.__class__.__name__)
        self.setStyleSheet(style)


class TextArea(QtWidgets.QTextEdit, Widget):
    def __init__(
        self,
        on_change: typing.Optional[typing.Callable] = None,
        style: typing.Optional[str] = "",
    ):
        super().__init__()

        if on_change:
            self.textChanged.connect(on_change)

        self.setObjectName(self.__class__.__name__)
        self.setStyleSheet(style)


@dataclass
class Option:
    label: str
    value: typing.Any


class Select(QtWidgets.QComboBox, Widget):
    def __init__(
        self,
        options: typing.Optional[typing.List[Option]] = None,
        on_change: typing.Optional[typing.Callable] = None,
        style: typing.Optional[str] = "",
    ):
        super().__init__()

        if options:
            self.addItems(list(map(lambda x: x.label, options)))

        if on_change:
            self.activated.connect(lambda idx: on_change(options[idx].value))

        # Set placeholder text for the combo box
        self.setEditable(True)
        self.lineEdit().setPlaceholderText("Select an option...")
        self.lineEdit().setAlignment(QtCore.Qt.AlignCenter)
        self.setEditable(False)  # Disable edit after setting placeholder

        # Set custom background color and styling using CSS
        # self.setStyleSheet(
        #     """
        #         QComboBox {
        #             background-color: #EFEFEF;
        #             border: 1px solid #AFAFAF;
        #             padding: 5px;
        #             border-radius: 5px;
        #         }
        #         QComboBox::drop-down {
        #             subcontrol-origin: padding;
        #             subcontrol-position: top right;
        #             width: 20px;
        #             border-left-width: 1px;
        #             border-left-color: darkgray;
        #             border-left-style: solid; /* Just a single line */
        #             border-top-right-radius: 3px; /* same radius as the QComboBox */
        #             border-bottom-right-radius: 3px;
        #         }
        #         QComboBox::down-arrow {
        #             image: url(icons/down_arrow.png);  /* Custom down-arrow icon */
        #         }
        #         QComboBox QAbstractItemView {
        #             background-color: white;
        #             selection-background-color: lightgray;
        #             selection-color: black;
        #         }
        #     """
        # )
        self.setStyleSheet(
            style
            + """
            Select::drop-down {
                subcontrol-origin: padding;
                subcontrol-position: top right;
                width: 5px;
                padding: 5px;
            }
            """
        )



class CSSProcessor:

    STYLESHEET_PATH: str
    STYLESHEET_TYPE: str

    STYLESHEET_TYPES = ["CSS"]

    def __init__(self, stylesheet_path: str, stylesheet_type="CSS"):
        super(CSSProcessor, self).__init__()

        # validate style sheet type (if in STYLESHEET_TYPES)

        if stylesheet_type not in self.STYLESHEET_TYPES:
            raise ValueError(
                "%s is not supported, available options are %s"
                % (stylesheet_type, self.STYLESHEET_TYPES)
            )

        self.STYLESHEET_PATH = stylesheet_path
        self.STYLESHEET_TYPE = stylesheet_type

    def trim_text(self, text):
        escapes = "".join([chr(char) for char in range(1, 15)])
        translator = str.maketrans("", "", escapes)

        return text.translate(translator)

    def read_style_sheet(self):
        style = ""

        if self.STYLESHEET_PATH:
            with open(self.STYLESHEET_PATH, "r") as file:
                for line in file.readlines():
                    # ignore comments
                    if not line.startswith("/*") and not line.endswith("*/"):
                        style += line

        return style

    def get(self, class_names: str = None, style: str = "") -> str:
        if class_names:
            __class_names: list = class_names.split(" ")

            def parse_class_name(class_name: str) -> str:
                class_block = self.read_style_sheet()[
                    self.read_style_sheet().find(".%s" % class_name)
                    + len(class_name)
                    + 1 :
                ]

                starts_at = class_block.find("{") + 1
                ends_at = class_block.find("}")

                return self.trim_text(class_block[starts_at:ends_at])

            return "".join(map(parse_class_name, __class_names)) + " " + style

        return style



import typing

from PyQt5 import QtWidgets

from melid.layout import Box


class Route(Box):
    _PATH: str
    _COMPONENT: str

    def __init__(self, path: str, component: QtWidgets.QWidget, **kwargs):
        super().__init__(child=component, **kwargs)

        self._PATH = path
        self._COMPONENT = component

    def navigate(self, path: str):
        Router.navigate(path)


class Router(Box):
    _INSTANCE = None
    _ROUTE_TREE: typing.Dict[str, int] = {}

    def __new__(cls, *args, **kwargs):
        if not cls._INSTANCE:
            cls._INSTANCE = super(Router, cls).__new__(cls, *args, **kwargs)
        return cls._INSTANCE

    def __init__(self, *routes: Route, **kwargs):
        super().__init__(**kwargs)

        self.router = QtWidgets.QStackedWidget()

        self.addWidget(self.router)
        self.addRoute(*routes)

        self.setCurrentIndex(self.router.currentIndex())

    def setCurrentIndex(self, index: int):
        self.router.setCurrentIndex(index)

    def addRoute(self, *routes: Route):
        for idx, route in enumerate(routes):
            self._ROUTE_TREE[route._PATH] = idx
            self.router.addWidget(route._COMPONENT)

    @classmethod
    def navigate(cls, path: str):
        if not cls._INSTANCE:
            raise ValueError("Router has not been initialized")

        idx = cls._ROUTE_TREE.get(path)
        if idx is None:
            raise ValueError("Page not found")

        cls._INSTANCE.setCurrentIndex(idx)



from typing import Any, Dict


def generate_size_utilities(sizes: Dict[str, str]) -> Dict[str, Any]:
    """
    Generate color utilities for padding, margin, border width, border radius, width and height
    """
    utilities = {}

    for name, value in sizes.items():
        # Padding Utilities
        utilities[f"p-{name}"] = {
            "declarations": [{"property": "padding", "value": value}]
        }
        utilities[f"px-{name}"] = {
            "declarations": [
                {"property": "padding-left", "value": value},
                {"property": "padding-right", "value": value},
            ]
        }
        utilities[f"py-{name}"] = {
            "declarations": [
                {"property": "padding-top", "value": value},
                {"property": "padding-bottom", "value": value},
            ],
        }
        utilities[f"pl-{name}"] = {
            "declarations": [{"property": "padding-left", "value": value}]
        }
        utilities[f"pr-{name}"] = {
            "declarations": [{"property": "padding-right", "value": value}]
        }
        utilities[f"pt-{name}"] = {
            "declarations": [{"property": "padding-top", "value": value}]
        }
        utilities[f"pb-{name}"] = {
            "declarations": [{"property": "padding-bottom", "value": value}]
        }

        # Margin Utilities
        utilities[f"m-{name}"] = {
            "declarations": [{"property": "margin", "value": value}]
        }
        utilities[f"mx-{name}"] = {
            "declarations": [
                {"property": "margin-left", "value": value},
                {"property": "margin-right", "value": value},
            ]
        }
        utilities[f"my-{name}"] = {
            "declarations": [
                {"property": "margin-top", "value": value},
                {"property": "margin-bottom", "value": value},
            ],
        }
        utilities[f"ml-{name}"] = {
            "declarations": [{"property": "margin-left", "value": value}]
        }
        utilities[f"mr-{name}"] = {
            "declarations": [{"property": "margin-right", "value": value}]
        }
        utilities[f"mt-{name}"] = {
            "declarations": [{"property": "margin-top", "value": value}]
        }
        utilities[f"mb-{name}"] = {
            "declarations": [{"property": "margin-bottom", "value": value}]
        }

        # Border Utilities
        utilities[f"border-{name}"] = {
            "declarations": [{"property": "border-width", "value": value}]
        }
        utilities[f"border-x-{name}"] = {
            "declarations": [
                {"property": "border-left-width", "value": value},
                {"property": "border-right-width", "value": value},
            ]
        }
        utilities[f"border-y-{name}"] = {
            "declarations": [
                {"property": "border-top-width", "value": value},
                {"property": "border-bottom-width", "value": value},
            ],
        }
        utilities[f"border-l-{name}"] = {
            "declarations": [{"property": "border-left-width", "value": value}]
        }
        utilities[f"border-r-{name}"] = {
            "declarations": [{"property": "border-right-width", "value": value}]
        }
        utilities[f"border-t-{name}"] = {
            "declarations": [{"property": "border-top-width", "value": value}]
        }
        utilities[f"border-b-{name}"] = {
            "declarations": [{"property": "border-bottom-width", "value": value}]
        }

        # Border Radius Utilities
        # TODO: direction combos (top-right, bottom-left, etc)
        utilities[f"rounded-{name}"] = {
            "declarations": [{"property": "border-radius", "value": value}]
        }
        utilities[f"rounded-l-{name}"] = {
            "declarations": [
                {"property": "border-top-left-radius", "value": value},
                {"property": "border-bottom-left-radius", "value": value},
            ]
        }
        utilities[f"rounded-r-{name}"] = {
            "declarations": [
                {"property": "border-top-right-radius", "value": value},
                {"property": "border-bottom-right-radius", "value": value},
            ]
        }
        utilities[f"rounded-t-{name}"] = {
            "declarations": [
                {"property": "border-top-left-radius", "value": value},
                {"property": "border-top-right-radius", "value": value},
            ]
        }
        utilities[f"rounded-b-{name}"] = {
            "declarations": [
                {"property": "border-bottom-left-radius", "value": value},
                {"property": "border-bottom-right-radius", "value": value},
            ]
        }

        # Width and Height Utilities
        utilities[f"w-{name}"] = {
            "declarations": [{"property": "width", "value": value}]
        }
        utilities[f"h-{name}"] = {
            "declarations": [{"property": "height", "value": value}]
        }
        utilities[f"min-w-{name}"] = {
            "declarations": [{"property": "min-width", "value": value}]
        }
        utilities[f"min-h-{name}"] = {
            "declarations": [{"property": "min-height", "value": value}]
        }
        utilities[f"max-w-{name}"] = {
            "declarations": [{"property": "max-width", "value": value}]
        }
        utilities[f"max-h-{name}"] = {
            "declarations": [{"property": "max-height", "value": value}]
        }

    return utilities



import os
import sys
import typing

from PyQt5 import QtCore, QtWidgets

from melid.types import AppEvents
from melid.utils import Size
from melid.widgets import Widget


class App(QtWidgets.QMainWindow):
    __QT_WINDOW_RESIZED = QtCore.pyqtSignal(AppEvents.WINDOW_RESIZED.value)

    def __init__(
        self,
        child: typing.Optional[
            typing.Union[QtWidgets.QWidget, typing.List[QtWidgets.QWidget]]
        ] = None,
        qt_app: typing.Optional[QtWidgets.QApplication] = None,
        name: typing.Optional[str] = "Melid",
        title: typing.Optional[str] = "Melid || Made with ❤ by Struckchure",
        size: typing.Optional[Size] = Size(minimum_width=500, minimum_height=500),
    ):
        super().__init__()

        self.__QT_APP = qt_app
        self.__QT_APP.setApplicationName(name)

        self.__central_widget = Widget(qt_window_resized=self.__QT_WINDOW_RESIZED)
        if child:
            if isinstance(child, list) and len(child) > 0:
                self.__central_widget.addWidgets(child)

            if isinstance(child, QtWidgets.QWidget):
                self.__central_widget.addWidget(child)

        size.apply(self.__central_widget).apply(self)

        self.setWindowTitle(title)
        self.setCentralWidget(self.__central_widget)
        self.setContentsMargins(0, 0, 0, 0)

    def addWidget(self, *args, **kwargs):
        return self.__central_widget.addWidget(*args, **kwargs)

    def center(self):
        cursor_position = QtWidgets.QApplication.desktop().cursor().pos()

        screen = QtWidgets.QApplication.screenAt(cursor_position)
        if not screen:
            screen = QtWidgets.QApplication.primaryScreen()

        if screen:
            screen_geometry = screen.geometry()
            screen_center = screen_geometry.center()

            # BUG: does not work as intended on all screens
            screen_center.setX(int(screen_center.x() - (screen_center.x() * 0.075)))
            screen_center.setY(int(screen_center.y() - (screen_center.y() * 0.5)))

            self.move(screen_center)

        return self

    def resizeEvent(self, a0):
        self.__QT_WINDOW_RESIZED.emit(a0.size())

        super().resizeEvent(a0)

    def showMaximized(self):
        super().showMaximized()

        return self

    def mount(self):
        self.show()

        sys.exit(self.__QT_APP.exec_())


os.environ["QT_MAC_WANTS_LAYER"] = "1"

qt_app = QtWidgets.QApplication(sys.argv)
qt_app.setQuitOnLastWindowClosed(True)


def CreateApp(*args, **kwargs):
    try:
        return App(qt_app=qt_app, *args, **kwargs)
    except Exception as e:
        print(str(e))

        sys.exit(qt_app.exec_())



import typing

from PyQt5 import QtCore


class Store:

    _instance = None
    _INITIAL_STATE: dict

    def __init__(self, initialState: dict[str, typing.Any] | None) -> None:
        if initialState:
            self._INITIAL_STATE = initialState

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls)

            cls.args = args
            cls.kwargs = kwargs

        return cls._instance

    def setState(self, state: dict):
        self._INITIAL_STATE = {**self.state, **state}

    @property
    def state(self) -> dict:
        return self._INITIAL_STATE


class State(QtCore.QObject):
    __value: typing.Any

    updated = QtCore.pyqtSignal(object)

    def __init__(self, value: typing.Any):
        super().__init__()

        self.__value = value

    def get(self):
        return self.__value

    def set(self, value):
        if self.__value != value:
            self.__value = value
            self.updated.emit(self.__value)

    def subscribe(self, callback: typing.Callable[[typing.Any], None]):
        self.updated.connect(callback)


class StatefulWidget:
    def state(
        self,
        state: State,
        condition: typing.Union[bool, typing.Callable[[], bool]] = True,
        attribute: str = None,
        params: typing.Union[typing.List, typing.Dict, typing.Callable] = None,
    ):
        def resolve_params(p):
            if callable(p):
                return resolve_params(p())

            p1 = p if isinstance(p, list) else []
            p2 = p if isinstance(p, dict) else {}

            return p1, p2

        def resolve_condition(c):
            if callable(c):
                return resolve_condition(c())

            return c

        def on_state_change():
            p1, p2 = resolve_params(params)

            if resolve_condition(condition) and attribute and hasattr(self, attribute):
                getattr(self, attribute)(*p1, **p2)

        state.subscribe(on_state_change)

        return self



import typing

from PyQt5 import QtCore, QtWidgets


class Size:
    def __init__(
        self,
        width: typing.Optional[int] = None,
        height: typing.Optional[int] = None,
        minimum_width: typing.Optional[int] = 0,
        minimum_height: typing.Optional[int] = 0,
        maximum_width: typing.Optional[int] = QtWidgets.QWIDGETSIZE_MAX,
        maximum_height: typing.Optional[int] = QtWidgets.QWIDGETSIZE_MAX,
    ):
        # Set the fixed size if width and height are provided
        self.fixed_size = QtCore.QSize(width or 0, height or 0)

        # Set minimum and maximum sizes
        self.minimum_size = QtCore.QSize(minimum_width, minimum_height)
        self.maximum_size = QtCore.QSize(maximum_width, maximum_height)

    def size(self) -> QtCore.QSize:
        """
        Returns the right size based on the PyQt5 preference for fixed, min, and max sizes.
        """
        # If a fixed size is set (both width and height)
        if self.fixed_size.width() > 0 and self.fixed_size.height() > 0:
            return self.fixed_size

        # Get the current size hint
        current_size = self.sizeHint()

        # Ensure the current size respects the min and max constraints
        width = max(
            self.minimum_size.width(),
            min(current_size.width(), self.maximum_size.width()),
        )
        height = max(
            self.minimum_size.height(),
            min(current_size.height(), self.maximum_size.height()),
        )

        return QtCore.QSize(width, height)

    def sizePolicy(self) -> QtWidgets.QSizePolicy:
        """
        Returns the appropriate size policies based on the fixed, minimum, and maximum sizes.
        """
        horizontal_policy = QtWidgets.QSizePolicy.Expanding
        vertical_policy = QtWidgets.QSizePolicy.Expanding

        # # Determine horizontal policy
        if self.fixed_size.width() > 0:
            horizontal_policy = QtWidgets.QSizePolicy.Fixed
        elif self.minimum_size.width() == self.maximum_size.width():
            horizontal_policy = QtWidgets.QSizePolicy.Fixed
        elif self.maximum_size.width() == QtWidgets.QWIDGETSIZE_MAX:
            horizontal_policy = QtWidgets.QSizePolicy.Expanding

        # Determine vertical policy
        if self.fixed_size.height() > 0:
            vertical_policy = QtWidgets.QSizePolicy.Fixed
        elif self.minimum_size.height() == self.maximum_size.height():
            vertical_policy = QtWidgets.QSizePolicy.Fixed
        elif self.maximum_size.height() == QtWidgets.QWIDGETSIZE_MAX:
            vertical_policy = QtWidgets.QSizePolicy.Expanding

        return QtWidgets.QSizePolicy(horizontal_policy, vertical_policy)

    def apply(self, widget: QtWidgets.QWidget):
        widget.setSizePolicy(self.sizePolicy())

        if self.fixed_size.width() > 0:
            widget.setFixedWidth(self.fixed_size.width())

        if self.fixed_size.height() > 0:
            widget.setFixedHeight(self.fixed_size.height())

        if self.minimum_size.width() > 0:
            widget.setMinimumWidth(self.minimum_size.width())

        if self.minimum_size.height() > 0:
            widget.setMinimumHeight(self.minimum_size.height())

        if self.maximum_size.width() > 0:
            widget.setMaximumWidth(self.maximum_size.width())

        if self.maximum_size.height() > 0:
            widget.setMaximumHeight(self.maximum_size.height())

        return self



import re
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Callable, Dict, List, Optional, Tuple, Union


@dataclass
class Rule:
    kind: str = field(init=False, default="rule")
    selector: str
    nodes: List["AstNode"]


@dataclass
class Declaration:
    kind: str = field(init=False, default="declaration")
    property: str
    value: Optional[str] = None
    important: bool = False


@dataclass
class Comment:
    kind: str = field(init=False, default="comment")
    value: str


@dataclass
class Context:
    kind: str = field(init=False, default="context")
    context: Dict[str, str]
    nodes: List["AstNode"]


@dataclass
class AtRoot:
    kind: str = field(init=False, default="at-root")
    nodes: List["AstNode"]


AstNode = Union[Rule, Declaration, Comment, Context, AtRoot]


class WalkAction(Enum):
    CONTINUE = "CONTINUE"
    SKIP = "SKIP"
    STOP = "STOP"


class Ast:
    @classmethod
    def walk(
        cls,
        ast: List[AstNode],
        visit: Callable[[AstNode, Dict[str, Any]], Optional[WalkAction]],
        parent: Optional[AstNode] = None,
        context: Optional[Dict[str, str]] = None,
    ) -> None:
        if context is None:
            context = {}

        i = 0
        while i < len(ast):
            node = ast[i]

            # Handle Context Nodes
            if isinstance(node, Context):
                new_context = {**context, **node.context}
                cls.walk(node.nodes, visit, parent, new_context)
                i += 1
                continue

            # Prepare utility functions
            utils = {
                "parent": parent,
                "replace_with": lambda new_node: cls.replace_with(ast, i, new_node),
                "context": context,
            }

            # Visit the node
            result = visit(node, utils)
            status = result if result is not None else WalkAction.CONTINUE

            # Handle WalkAction
            if status == WalkAction.STOP:
                return
            elif status == WalkAction.SKIP:
                i += 1
                continue

            # Recursively walk child nodes if it's a Rule
            if isinstance(node, Rule):
                cls.walk(node.nodes, visit, node, context)

            i += 1

    def replace_with(
        self,
        ast: List[AstNode],
        index: int,
        new_node: Union[AstNode, List[AstNode]],
    ) -> None:
        if isinstance(new_node, list):
            ast[index : index + 1] = new_node
        else:
            ast[index : index + 1] = [new_node]
        # Decrement index to revisit the replaced node(s)
        # This is handled in the walk function by not incrementing 'i' after replacement

    @classmethod
    def rule(cls, selector: str, nodes: List[AstNode]) -> Rule:
        return Rule(selector=selector, nodes=nodes)

    @classmethod
    def decl(
        cls, property: str, value: Optional[str] = None, important: bool = False
    ) -> Declaration:
        return Declaration(property=property, value=value, important=important)

    @classmethod
    def comment(cls, value: str) -> Comment:
        return Comment(value=value)

    @classmethod
    def context_node(cls, context: Dict[str, str], nodes: List[AstNode]) -> Context:
        return Context(context=context, nodes=nodes)

    @classmethod
    def at_root(cls, nodes: List[AstNode]) -> AtRoot:
        return AtRoot(nodes=nodes)

    def parse_stylesheet(self, css: str) -> List[AstNode]:
        """
        Parse a CSS stylesheet string and convert it into an AST.
        """
        tokens = self.tokenize(css)
        ast, _ = self.parse_rules(tokens, 0)
        return ast

    def tokenize(self, css: str) -> List[str]:
        """
        Tokenize a CSS string into individual tokens (selectors, properties, values, etc.).
        """
        # Remove comments and extra whitespace
        css = re.sub(r"/\*[^*]*\*+([^/*][^*]*\*+)*/", "", css)
        css = css.strip()

        # Tokenize using regex (captures braces, semicolons, colons, etc.)
        token_pattern = re.compile(r"([{};:])|\s+")
        tokens = [
            token for token in token_pattern.split(css) if token and not token.isspace()
        ]
        return tokens

    def parse_rules(self, tokens: List[str], pos: int) -> Tuple[List[AstNode], int]:
        """
        Parse CSS tokens and convert them into a list of Rule AST nodes.
        """
        ast_nodes = []
        current_rule = None
        current_declarations = []

        while pos < len(tokens):
            token = tokens[pos]

            if token == "{":
                # Start of declaration block
                current_rule.nodes, pos = self.parse_declarations(tokens, pos + 1)
                ast_nodes.append(current_rule)
                current_rule = None

            elif token == "}":
                # End of block
                break

            elif token == ";":
                # End of a declaration
                if current_declarations:
                    ast_nodes.append(
                        self.decl(current_declarations[0], current_declarations[1])
                    )
                current_declarations = []

            elif ":" in token:
                # It's a property declaration
                property_name = tokens[pos - 1].strip()
                value = tokens[pos + 1].strip()
                current_declarations = [property_name, value]
                pos += 2  # Move past value

            else:
                # Assume it's a selector if not inside a block
                if current_rule is None:
                    current_rule = self.rule(selector=token.strip(), nodes=[])
            pos += 1

        return ast_nodes, pos

    def parse_declarations(
        self, tokens: List[str], pos: int
    ) -> Tuple[List[AstNode], int]:
        """
        Parse a block of declarations and return as Declaration AST nodes.
        """
        declarations = []
        while pos < len(tokens):
            token = tokens[pos]

            if token == "}":
                break  # End of block

            elif ":" in token:
                # Parse a declaration
                property_name = tokens[pos - 1].strip()
                value = tokens[pos + 1].strip()
                declarations.append(self.decl(property_name, value))
                pos += 2  # Move past the value
            elif token == ";":
                pos += 1  # Move past semicolon
            else:
                pos += 1  # Skip over unknown tokens

        return declarations, pos


# # Example usage
# css_string = """
#     .container {
#         width: 100%;
#         padding: 1rem;
#     }

#     .button {
#         background-color: blue;
#         border-radius: 0.5rem;
#     }
# """

# ast = parse_stylesheet(css_string)
# print(ast)


# __initial_ast: List[AstNode] = [
#     rule(
#         "body",
#         [
#             decl("background-color", "white"),
#             decl("color", "black"),
#             comment("This is a comment"),
#         ],
#     ),
#     rule(
#         ".container",
#         [
#             decl("max-width", "1200px"),
#             decl("margin", "0 auto"),
#         ],
#     ),
#     context_node(
#         {"media": "screen"},
#         [
#             rule(
#                 "@media screen and (max-width: 600px)",
#                 [
#                     decl("background-color", "lightgray"),
#                 ],
#             ),
#         ],
#     ),
#     at_root(
#         [
#             rule(
#                 ":root",
#                 [
#                     decl("--main-color", "#333"),
#                 ],
#             )
#         ]
#     ),
#     rule(
#         "@property --custom-color",
#         [
#             decl("initial-value", "#fff"),
#             decl("inherits", "true"),
#         ],
#     ),
# ]
